//=======================================================================
// Copyright 2014-2015 David Simmons-Duffin.
// Distributed under the MIT License.
// (See accompanying file LICENSE or copy at
//  http://opensource.org/licenses/MIT)
//=======================================================================


#ifndef SDPB_TYPES_H_
#define SDPB_TYPES_H_

#include <mblas.h>
#include <mlapack.h>

typedef mpackint Integer;
typedef mpf_class Real;

#endif  // SDPB_TYPES_H_
